import inspect
import logging
from collections import defaultdict
from contextlib import ExitStack
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import pendulum
from typing_extensions import TypeAlias

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.asset_check_evaluation import AssetCheckEvaluation
from dagster._core.definitions.events import (
    AssetMaterialization,
    AssetObservation,
)
from dagster._core.definitions.instigation_logger import InstigationLogger
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.partition import (
    CachingDynamicPartitionsLoader,
)
from dagster._core.definitions.resource_annotation import (
    get_resource_args,
)
from dagster._core.definitions.resource_definition import (
    Resources,
)
from dagster._core.definitions.scoped_resources_builder import ScopedResourcesBuilder
from dagster._core.errors import (
    DagsterInvalidDefinitionError,
    DagsterInvalidInvocationError,
    DagsterInvalidSubsetError,
    DagsterInvariantViolationError,
)
from dagster._core.instance import DagsterInstance
from dagster._core.instance.ref import InstanceRef
from dagster._serdes import whitelist_for_serdes
from dagster._utils import IHasInternalInit, normalize_to_repository

from ..decorator_utils import (
    get_function_params,
)
from .asset_selection import AssetSelection
from .graph_definition import GraphDefinition
from .run_request import (
    AddDynamicPartitionsRequest,
    DagsterRunReaction,
    DeleteDynamicPartitionsRequest,
    RunRequest,
    SensorResult,
    SkipReason,
)
from .target import DirectTarget, ExecutableDefinition, RepoRelativeTarget
from .unresolved_asset_job_definition import UnresolvedAssetJobDefinition
from .utils import check_valid_name

if TYPE_CHECKING:
    from dagster import ResourceDefinition
    from dagster._core.definitions.definitions_class import Definitions
    from dagster._core.definitions.repository_definition import RepositoryDefinition


@whitelist_for_serdes
class DefaultSensorStatus(Enum):
    RUNNING = "RUNNING"
    STOPPED = "STOPPED"


@whitelist_for_serdes
class SensorType(Enum):
    STANDARD = "STANDARD"
    RUN_STATUS = "RUN_STATUS"
    ASSET = "ASSET"
    MULTI_ASSET = "MULTI_ASSET"
    FRESHNESS_POLICY = "FRESHNESS_POLICY"
    UNKNOWN = "UNKNOWN"


DEFAULT_SENSOR_DAEMON_INTERVAL = 30


class SensorEvaluationContext:
    """The context object available as the argument to the evaluation function of a :py:class:`dagster.SensorDefinition`.

    Users should not instantiate this object directly. To construct a
    `SensorEvaluationContext` for testing purposes, use :py:func:`dagster.
    build_sensor_context`.

    Attributes:
        instance_ref (Optional[InstanceRef]): The serialized instance configured to run the schedule
        cursor (Optional[str]): The cursor, passed back from the last sensor evaluation via
            the cursor attribute of SkipReason and RunRequest
        last_completion_time (float): DEPRECATED The last time that the sensor was evaluated (UTC).
        last_run_key (str): DEPRECATED The run key of the RunRequest most recently created by this
            sensor. Use the preferred `cursor` attribute instead.
        repository_name (Optional[str]): The name of the repository that the sensor belongs to.
        repository_def (Optional[RepositoryDefinition]): The repository or that
            the sensor belongs to. If needed by the sensor top-level resource definitions will be
            pulled from this repository. You can provide either this or `definitions`.
        instance (Optional[DagsterInstance]): The deserialized instance can also be passed in
            directly (primarily useful in testing contexts).
        definitions (Optional[Definitions]): `Definitions` object that the sensor is defined in.
            If needed by the sensor, top-level resource definitions will be pulled from these
            definitions. You can provide either this or `repository_def`.
        resources (Optional[Dict[str, Any]]): A dict of resource keys to resource
            definitions to be made available during sensor execution.

    Example:
        .. code-block:: python

            from dagster import sensor, SensorEvaluationContext

            @sensor
            def the_sensor(context: SensorEvaluationContext):
                ...

    """

    def __init__(
        self,
        instance_ref: Optional[InstanceRef],
        last_completion_time: Optional[float],
        last_run_key: Optional[str],
        cursor: Optional[str],
        repository_name: Optional[str],
        repository_def: Optional["RepositoryDefinition"] = None,
        instance: Optional[DagsterInstance] = None,
        sensor_name: Optional[str] = None,
        resources: Optional[Mapping[str, "ResourceDefinition"]] = None,
        definitions: Optional["Definitions"] = None,
    ):
        from dagster._core.definitions.definitions_class import Definitions
        from dagster._core.definitions.repository_definition import RepositoryDefinition

        self._exit_stack = ExitStack()
        self._instance_ref = check.opt_inst_param(instance_ref, "instance_ref", InstanceRef)
        self._last_completion_time = check.opt_float_param(
            last_completion_time, "last_completion_time"
        )
        self._last_run_key = check.opt_str_param(last_run_key, "last_run_key")
        self._cursor = check.opt_str_param(cursor, "cursor")
        self._repository_name = check.opt_str_param(repository_name, "repository_name")
        self._repository_def = normalize_to_repository(
            check.opt_inst_param(definitions, "definitions", Definitions),
            check.opt_inst_param(repository_def, "repository_def", RepositoryDefinition),
            error_on_none=False,
        )
        self._instance = check.opt_inst_param(instance, "instance", DagsterInstance)
        self._sensor_name = sensor_name

        # Wait to set resources unless they're accessed
        self._resource_defs = resources
        self._resources = None
        self._cm_scope_entered = False

        self._log_key = (
            [
                repository_name,
                sensor_name,
                pendulum.now("UTC").strftime("%Y%m%d_%H%M%S"),
            ]
            if repository_name and sensor_name
            else None
        )
        self._logger: Optional[InstigationLogger] = None
        self._cursor_updated = False

    def __enter__(self) -> "SensorEvaluationContext":
        self._cm_scope_entered = True
        return self

    def __exit__(self, *exc) -> None:
        self._exit_stack.close()
        self._logger = None

    @property
    def resource_defs(self) -> Optional[Mapping[str, "ResourceDefinition"]]:
        return self._resource_defs

    @property
    def sensor_name(self) -> str:
        return check.not_none(self._sensor_name, "Only valid when sensor name provided")

    def merge_resources(self, resources_dict: Mapping[str, Any]) -> "SensorEvaluationContext":
        """Merge the specified resources into this context.

        This method is intended to be used by the Dagster framework, and should not be called by user code.

        Args:
            resources_dict (Mapping[str, Any]): The resources to replace in the context.
        """
        check.invariant(
            self._resources is None, "Cannot merge resources in context that has been initialized."
        )
        from dagster._core.execution.build_resources import wrap_resources_for_execution

        return SensorEvaluationContext(
            instance_ref=self._instance_ref,
            last_completion_time=self._last_completion_time,
            last_run_key=self._last_run_key,
            cursor=self._cursor,
            repository_name=self._repository_name,
            repository_def=self._repository_def,
            instance=self._instance,
            sensor_name=self._sensor_name,
            resources={
                **(self._resource_defs or {}),
                **wrap_resources_for_execution(resources_dict),
            },
        )

    @public
    @property
    def resources(self) -> Resources:
        """Resources: A mapping from resource key to instantiated resources for this sensor."""
        from dagster._core.definitions.scoped_resources_builder import (
            IContainsGenerator,
        )
        from dagster._core.execution.build_resources import build_resources

        if not self._resources:
            """
            This is similar to what we do in e.g. the op context - we set up a resource
            building context manager, and immediately enter it. This is so that in cases
            where a user is not using any context-manager based resources, they don't
            need to enter this SensorEvaluationContext themselves.

            For example:

            my_sensor(build_sensor_context(resources={"my_resource": my_non_cm_resource})

            will work ok, but for a CM resource we must do

            with build_sensor_context(resources={"my_resource": my_cm_resource}) as context:
                my_sensor(context)
            """

            # Early exit if no resources are defined. This skips unnecessary initialization
            # entirely. This allows users to run user code servers in cases where they
            # do not have access to the instance if they use a subset of features do
            # that do not require instance access. In this case, if they do not use
            # resources on sensors they do not require the instance, so we do not
            # instantiate it
            #
            # Tracking at https://github.com/dagster-io/dagster/issues/14345
            if not self._resource_defs:
                self._resources = ScopedResourcesBuilder.build_empty()
                return self._resources

            instance = self.instance if self._instance or self._instance_ref else None

            resources_cm = build_resources(resources=self._resource_defs or {}, instance=instance)
            self._resources = self._exit_stack.enter_context(resources_cm)

            if isinstance(self._resources, IContainsGenerator) and not self._cm_scope_entered:
                self._exit_stack.close()
                raise DagsterInvariantViolationError(
                    "At least one provided resource is a generator, but attempting to access"
                    " resources outside of context manager scope. You can use the following syntax"
                    " to open a context manager: `with build_schedule_context(...) as context:`"
                )

        return self._resources

    @public
    @property
    def instance(self) -> DagsterInstance:
        """DagsterInstance: The current DagsterInstance."""
        # self._instance_ref should only ever be None when this SensorEvaluationContext was
        # constructed under test.
        if not self._instance:
            if not self._instance_ref:
                raise DagsterInvariantViolationError(
                    "Attempted to initialize dagster instance, but no instance reference was"
                    " provided."
                )
            self._instance = self._exit_stack.enter_context(
                DagsterInstance.from_ref(self._instance_ref)
            )
        return cast(DagsterInstance, self._instance)

    @property
    def instance_ref(self) -> Optional[InstanceRef]:
        return self._instance_ref

    @public
    @property
    def last_completion_time(self) -> Optional[float]:
        """Optional[float]: Timestamp representing the last time this sensor completed an evaluation."""
        return self._last_completion_time

    @public
    @property
    def last_run_key(self) -> Optional[str]:
        """Optional[str]: The run key supplied to the most recent RunRequest produced by this sensor."""
        return self._last_run_key

    @public
    @property
    def cursor(self) -> Optional[str]:
        """The cursor value for this sensor, which was set in an earlier sensor evaluation."""
        return self._cursor

    @public
    def update_cursor(self, cursor: Optional[str]) -> None:
        """Updates the cursor value for this sensor, which will be provided on the context for the
        next sensor evaluation.

        This can be used to keep track of progress and avoid duplicate work across sensor
        evaluations.

        Args:
            cursor (Optional[str]):
        """
        self._cursor = check.opt_str_param(cursor, "cursor")
        self._cursor_updated = True

    @property
    def cursor_updated(self) -> bool:
        return self._cursor_updated

    @public
    @property
    def repository_name(self) -> Optional[str]:
        """Optional[str]: The name of the repository that this sensor resides in."""
        return self._repository_name

    @public
    @property
    def repository_def(self) -> Optional["RepositoryDefinition"]:
        """Optional[RepositoryDefinition]: The RepositoryDefinition that this sensor resides in."""
        return self._repository_def

    @property
    def log(self) -> logging.Logger:
        if self._logger:
            return self._logger

        if not self._instance_ref:
            self._logger = self._exit_stack.enter_context(
                InstigationLogger(
                    self._log_key,
                    repository_name=self._repository_name,
                    name=self._sensor_name,
                )
            )
            return cast(logging.Logger, self._logger)

        self._logger = self._exit_stack.enter_context(
            InstigationLogger(
                self._log_key,
                self.instance,
                repository_name=self._repository_name,
                name=self._sensor_name,
            )
        )
        return cast(logging.Logger, self._logger)

    def has_captured_logs(self):
        return self._logger and self._logger.has_captured_logs()

    @property
    def log_key(self) -> Optional[List[str]]:
        return self._log_key


RawSensorEvaluationFunctionReturn = Union[
    Iterator[Union[SkipReason, RunRequest, DagsterRunReaction, SensorResult]],
    Sequence[RunRequest],
    SkipReason,
    RunRequest,
    DagsterRunReaction,
    SensorResult,
]
RawSensorEvaluationFunction: TypeAlias = Callable[..., RawSensorEvaluationFunctionReturn]

SensorEvaluationFunction: TypeAlias = Callable[..., Sequence[Union[SkipReason, RunRequest]]]


def get_context_param_name(fn: Callable) -> Optional[str]:
    """Determines the sensor's context parameter name by excluding all resource parameters."""
    resource_params = {param.name for param in get_resource_args(fn)}

    return next(
        (param.name for param in get_function_params(fn) if param.name not in resource_params), None
    )


def validate_and_get_resource_dict(
    resources: Resources, sensor_name: str, required_resource_keys: Set[str]
) -> Dict[str, Any]:
    """Validates that the context has all the required resources and returns a dictionary of
    resource key to resource object.
    """
    for k in required_resource_keys:
        if not hasattr(resources, k):
            raise DagsterInvalidDefinitionError(
                f"Resource with key '{k}' required by sensor '{sensor_name}' was not provided."
            )

    return {k: getattr(resources, k) for k in required_resource_keys}


def _check_dynamic_partitions_requests(
    dynamic_partitions_requests: Sequence[
        Union[AddDynamicPartitionsRequest, DeleteDynamicPartitionsRequest]
    ],
) -> None:
    req_keys_to_add_by_partitions_def_name = defaultdict(set)
    req_keys_to_delete_by_partitions_def_name = defaultdict(set)

    for req in dynamic_partitions_requests:
        duplicate_req_keys_to_delete = req_keys_to_delete_by_partitions_def_name.get(
            req.partitions_def_name, set()
        ).intersection(req.partition_keys)
        duplicate_req_keys_to_add = req_keys_to_add_by_partitions_def_name.get(
            req.partitions_def_name, set()
        ).intersection(req.partition_keys)
        if isinstance(req, AddDynamicPartitionsRequest):
            if duplicate_req_keys_to_delete:
                raise DagsterInvariantViolationError(
                    "Dynamic partition requests cannot contain both add and delete requests for"
                    " the same partition keys.Invalid request: partitions_def_name"
                    f" '{req.partitions_def_name}', partition_keys: {duplicate_req_keys_to_delete}"
                )
            elif duplicate_req_keys_to_add:
                raise DagsterInvariantViolationError(
                    "Cannot request to add duplicate dynamic partition keys: \npartitions_def_name"
                    f" '{req.partitions_def_name}', partition_keys: {duplicate_req_keys_to_add}"
                )
            req_keys_to_add_by_partitions_def_name[req.partitions_def_name].update(
                req.partition_keys
            )
        elif isinstance(req, DeleteDynamicPartitionsRequest):
            if duplicate_req_keys_to_delete:
                raise DagsterInvariantViolationError(
                    "Cannot request to add duplicate dynamic partition keys: \npartitions_def_name"
                    f" '{req.partitions_def_name}', partition_keys:"
                    f" {req_keys_to_add_by_partitions_def_name}"
                )
            elif duplicate_req_keys_to_add:
                raise DagsterInvariantViolationError(
                    "Dynamic partition requests cannot contain both add and delete requests for"
                    " the same partition keys.Invalid request: partitions_def_name"
                    f" '{req.partitions_def_name}', partition_keys: {duplicate_req_keys_to_add}"
                )
            req_keys_to_delete_by_partitions_def_name[req.partitions_def_name].update(
                req.partition_keys
            )
        else:
            check.failed(f"Unexpected dynamic partition request type: {req}")


class SensorDefinition(IHasInternalInit):
    """Define a sensor that initiates a set of runs based on some external state.

    Args:
        evaluation_fn (Callable[[SensorEvaluationContext]]): The core evaluation function for the
            sensor, which is run at an interval to determine whether a run should be launched or
            not. Takes a :py:class:`~dagster.SensorEvaluationContext`.

            This function must return a generator, which must yield either a single SkipReason
            or one or more RunRequest objects.
        name (Optional[str]): The name of the sensor to create. Defaults to name of evaluation_fn
        minimum_interval_seconds (Optional[int]): The minimum number of seconds that will elapse
            between sensor evaluations.
        description (Optional[str]): A human-readable description of the sensor.
        job (Optional[GraphDefinition, JobDefinition, UnresolvedAssetJob]): The job to execute when this sensor fires.
        jobs (Optional[Sequence[GraphDefinition, JobDefinition, UnresolvedAssetJob]]): (experimental) A list of jobs to execute when this sensor fires.
        default_status (DefaultSensorStatus): Whether the sensor starts as running or not. The default
            status can be overridden from the Dagster UI or via the GraphQL API.
        asset_selection (AssetSelection): (Experimental) an asset selection to launch a run for if
            the sensor condition is met. This can be provided instead of specifying a job.
    """

    def with_updated_jobs(self, new_jobs: Sequence[ExecutableDefinition]) -> "SensorDefinition":
        """Returns a copy of this sensor with the jobs replaced.

        Args:
            job (ExecutableDefinition): The job that should execute when this
                schedule runs.
        """
        return SensorDefinition.dagster_internal_init(
            name=self.name,
            evaluation_fn=self._raw_fn,
            minimum_interval_seconds=self.minimum_interval_seconds,
            description=self.description,
            job_name=None,  # if original init was passed job name, was resolved to a job
            jobs=new_jobs if len(new_jobs) > 1 else None,
            job=new_jobs[0] if len(new_jobs) == 1 else None,
            default_status=self.default_status,
            asset_selection=self.asset_selection,
            required_resource_keys=self._raw_required_resource_keys,
        )

    def with_updated_job(self, new_job: ExecutableDefinition) -> "SensorDefinition":
        """Returns a copy of this sensor with the job replaced.

        Args:
            job (ExecutableDefinition): The job that should execute when this
                schedule runs.
        """
        return self.with_updated_jobs([new_job])

    def __init__(
        self,
        name: Optional[str] = None,
        *,
        evaluation_fn: Optional[RawSensorEvaluationFunction] = None,
        job_name: Optional[str] = None,
        minimum_interval_seconds: Optional[int] = None,
        description: Optional[str] = None,
        job: Optional[ExecutableDefinition] = None,
        jobs: Optional[Sequence[ExecutableDefinition]] = None,
        default_status: DefaultSensorStatus = DefaultSensorStatus.STOPPED,
        asset_selection: Optional[AssetSelection] = None,
        required_resource_keys: Optional[Set[str]] = None,
    ):
        from dagster._config.pythonic_config import validate_resource_annotated_function

        if evaluation_fn is None:
            raise DagsterInvalidDefinitionError("Must provide evaluation_fn to SensorDefinition.")

        if (
            sum(
                [
                    int(job is not None),
                    int(jobs is not None),
                    int(job_name is not None),
                    int(asset_selection is not None),
                ]
            )
            > 1
        ):
            raise DagsterInvalidDefinitionError(
                "Attempted to provide more than one of 'job', 'jobs', 'job_name', and "
                "'asset_selection' params to SensorDefinition. Must provide only one."
            )

        jobs = jobs if jobs else [job] if job else None

        targets: Optional[List[Union[RepoRelativeTarget, DirectTarget]]] = None
        if job_name:
            targets = [
                RepoRelativeTarget(
                    job_name=check.str_param(job_name, "job_name"),
                    op_selection=None,
                )
            ]
        elif job:
            targets = [DirectTarget(job)]
        elif jobs:
            targets = [DirectTarget(job) for job in jobs]
        elif asset_selection:
            targets = []

        if name:
            self._name = check_valid_name(name)
        else:
            self._name = evaluation_fn.__name__

        self._raw_fn: RawSensorEvaluationFunction = check.callable_param(
            evaluation_fn, "evaluation_fn"
        )
        self._evaluation_fn: Union[
            SensorEvaluationFunction,
            Callable[
                [SensorEvaluationContext],
                List[Union[SkipReason, RunRequest, DagsterRunReaction]],
            ],
        ] = wrap_sensor_evaluation(self._name, evaluation_fn)
        self._min_interval = check.opt_int_param(
            minimum_interval_seconds, "minimum_interval_seconds", DEFAULT_SENSOR_DAEMON_INTERVAL
        )
        self._description = check.opt_str_param(description, "description")
        self._targets: Sequence[Union[RepoRelativeTarget, DirectTarget]] = check.opt_list_param(
            targets, "targets", (DirectTarget, RepoRelativeTarget)
        )
        self._default_status = check.inst_param(
            default_status, "default_status", DefaultSensorStatus
        )
        self._asset_selection = check.opt_inst_param(
            asset_selection, "asset_selection", AssetSelection
        )
        validate_resource_annotated_function(self._raw_fn)
        resource_arg_names: Set[str] = {arg.name for arg in get_resource_args(self._raw_fn)}

        check.param_invariant(
            len(required_resource_keys or []) == 0 or len(resource_arg_names) == 0,
            "Cannot specify resource requirements in both @sensor decorator and as arguments to"
            " the decorated function",
        )
        self._raw_required_resource_keys = check.opt_set_param(
            required_resource_keys, "required_resource_keys", of_type=str
        )
        self._required_resource_keys = self._raw_required_resource_keys or resource_arg_names

    @staticmethod
    def dagster_internal_init(
        *,
        name: Optional[str],
        evaluation_fn: Optional[RawSensorEvaluationFunction],
        job_name: Optional[str],
        minimum_interval_seconds: Optional[int],
        description: Optional[str],
        job: Optional[ExecutableDefinition],
        jobs: Optional[Sequence[ExecutableDefinition]],
        default_status: DefaultSensorStatus,
        asset_selection: Optional[AssetSelection],
        required_resource_keys: Optional[Set[str]],
    ) -> "SensorDefinition":
        return SensorDefinition(
            name=name,
            evaluation_fn=evaluation_fn,
            job_name=job_name,
            minimum_interval_seconds=minimum_interval_seconds,
            description=description,
            job=job,
            jobs=jobs,
            default_status=default_status,
            asset_selection=asset_selection,
            required_resource_keys=required_resource_keys,
        )

    def __call__(self, *args, **kwargs) -> RawSensorEvaluationFunctionReturn:
        context_param_name_if_present = get_context_param_name(self._raw_fn)
        context = get_or_create_sensor_context(self._raw_fn, *args, **kwargs)

        context_param = (
            {context_param_name_if_present: context} if context_param_name_if_present else {}
        )

        resources = validate_and_get_resource_dict(
            context.resources, self.name, self._required_resource_keys
        )
        return self._raw_fn(**context_param, **resources)

    @public
    @property
    def required_resource_keys(self) -> Set[str]:
        """Set[str]: The set of keys for resources that must be provided to this sensor."""
        return self._required_resource_keys

    @public
    @property
    def name(self) -> str:
        """str: The name of this sensor."""
        return self._name

    @public
    @property
    def description(self) -> Optional[str]:
        """Optional[str]: A description for this sensor."""
        return self._description

    @public
    @property
    def minimum_interval_seconds(self) -> Optional[int]:
        """Optional[int]: The minimum number of seconds between sequential evaluations of this sensor."""
        return self._min_interval

    @property
    def targets(self) -> Sequence[Union[DirectTarget, RepoRelativeTarget]]:
        return self._targets

    @public
    @property
    def job(self) -> Union[JobDefinition, GraphDefinition, UnresolvedAssetJobDefinition]:
        """Union[GraphDefinition, JobDefinition, UnresolvedAssetJobDefinition]: The job that is
        targeted by this schedule.
        """
        if self._targets:
            if len(self._targets) == 1 and isinstance(self._targets[0], DirectTarget):
                return self._targets[0].target
            elif len(self._targets) > 1:
                raise DagsterInvalidDefinitionError(
                    "Job property not available when SensorDefinition has multiple jobs."
                )
        raise DagsterInvalidDefinitionError("No job was provided to SensorDefinition.")

    @public
    @property
    def jobs(self) -> List[Union[JobDefinition, GraphDefinition, UnresolvedAssetJobDefinition]]:
        """List[Union[GraphDefinition, JobDefinition, UnresolvedAssetJobDefinition]]: A list of jobs
        that are targeted by this schedule.
        """
        if self._targets and all(isinstance(target, DirectTarget) for target in self._targets):
            return [target.target for target in self._targets]  # type: ignore  # (illegible conditional)
        raise DagsterInvalidDefinitionError("No job was provided to SensorDefinition.")

    @property
    def sensor_type(self) -> SensorType:
        return SensorType.STANDARD

    def evaluate_tick(self, context: "SensorEvaluationContext") -> "SensorExecutionData":
        """Evaluate sensor using the provided context.

        Args:
            context (SensorEvaluationContext): The context with which to evaluate this sensor.

        Returns:
            SensorExecutionData: Contains list of run requests, or skip message if present.

        """
        context = check.inst_param(context, "context", SensorEvaluationContext)

        result = self._evaluation_fn(context)

        skip_message: Optional[str] = None
        run_requests: List[RunRequest] = []
        dagster_run_reactions: List[DagsterRunReaction] = []
        dynamic_partitions_requests: Optional[
            Sequence[Union[AddDynamicPartitionsRequest, DeleteDynamicPartitionsRequest]]
        ] = []
        updated_cursor = context.cursor
        asset_events = []

        if not result or result == [None]:
            skip_message = "Sensor function returned an empty result"
        elif len(result) == 1:
            item = result[0]
            check.inst(item, (SkipReason, RunRequest, DagsterRunReaction, SensorResult))

            if isinstance(item, SensorResult):
                run_requests = list(item.run_requests) if item.run_requests else []
                skip_message = (
                    item.skip_reason.skip_message
                    if item.skip_reason
                    else (None if run_requests else "Sensor function returned an empty result")
                )

                _check_dynamic_partitions_requests(
                    item.dynamic_partitions_requests or [],
                )
                dynamic_partitions_requests = item.dynamic_partitions_requests or []

                if item.cursor and context.cursor_updated:
                    raise DagsterInvariantViolationError(
                        "SensorResult.cursor cannot be set if context.update_cursor() was called."
                    )
                updated_cursor = item.cursor
                asset_events = item.asset_events

            elif isinstance(item, RunRequest):
                run_requests = [item]
            elif isinstance(item, SkipReason):
                skip_message = item.skip_message if isinstance(item, SkipReason) else None
            elif isinstance(item, DagsterRunReaction):
                dagster_run_reactions = (
                    [cast(DagsterRunReaction, item)] if isinstance(item, DagsterRunReaction) else []
                )
            else:
                check.failed(f"Unexpected type {type(item)} in sensor result")
        else:
            if any(isinstance(item, SensorResult) for item in result):
                check.failed(
                    "When a SensorResult is returned from a sensor, it must be the only object"
                    " returned."
                )

            check.is_list(result, (SkipReason, RunRequest, DagsterRunReaction))
            has_skip = any(map(lambda x: isinstance(x, SkipReason), result))
            run_requests = [item for item in result if isinstance(item, RunRequest)]
            dagster_run_reactions = [
                item for item in result if isinstance(item, DagsterRunReaction)
            ]

            if has_skip:
                if len(run_requests) > 0:
                    check.failed(
                        "Expected a single SkipReason or one or more RunRequests: received both "
                        "RunRequest and SkipReason"
                    )
                elif len(dagster_run_reactions) > 0:
                    check.failed(
                        "Expected a single SkipReason or one or more DagsterRunReaction: "
                        "received both DagsterRunReaction and SkipReason"
                    )
                else:
                    check.failed("Expected a single SkipReason: received multiple SkipReasons")

        _check_dynamic_partitions_requests(dynamic_partitions_requests)
        resolved_run_requests = self.resolve_run_requests(
            run_requests, context, self._asset_selection, dynamic_partitions_requests
        )

        return SensorExecutionData(
            resolved_run_requests,
            skip_message,
            updated_cursor,
            dagster_run_reactions,
            captured_log_key=context.log_key if context.has_captured_logs() else None,
            dynamic_partitions_requests=dynamic_partitions_requests,
            asset_events=asset_events,
        )

    def has_loadable_targets(self) -> bool:
        for target in self._targets:
            if isinstance(target, DirectTarget):
                return True
        return False

    def load_targets(
        self,
    ) -> Sequence[Union[JobDefinition, GraphDefinition, UnresolvedAssetJobDefinition]]:
        """Returns job/graph definitions that have been directly passed into the sensor definition.
        Any jobs or graphs that are referenced by name will not be loaded.
        """
        targets = []
        for target in self._targets:
            if isinstance(target, DirectTarget):
                targets.append(target.load())
        return targets

    def resolve_run_requests(
        self,
        run_requests: Sequence[RunRequest],
        context: SensorEvaluationContext,
        asset_selection: Optional[AssetSelection],
        dynamic_partitions_requests: Sequence[
            Union[AddDynamicPartitionsRequest, DeleteDynamicPartitionsRequest]
        ],
    ) -> Sequence[RunRequest]:
        def _get_repo_job_by_name(context: SensorEvaluationContext, job_name: str) -> JobDefinition:
            if context.repository_def is None:
                raise DagsterInvariantViolationError(
                    "Must provide repository def to build_sensor_context when yielding partitioned"
                    " run requests"
                )
            return context.repository_def.get_job(job_name)

        has_multiple_targets = len(self._targets) > 1
        target_names = [target.job_name for target in self._targets]

        if run_requests and len(self._targets) == 0 and not self._asset_selection:
            raise Exception(
                f"Error in sensor {self._name}: Sensor evaluation function returned a RunRequest "
                "for a sensor lacking a specified target (job_name, job, or jobs). Targets "
                "can be specified by providing job, jobs, or job_name to the @sensor "
                "decorator."
            )

        if asset_selection:
            run_requests = [
                *_run_requests_with_base_asset_jobs(run_requests, context, asset_selection)
            ]

        dynamic_partitions_store = (
            CachingDynamicPartitionsLoader(context.instance) if context.instance_ref else None
        )

        # Run requests may contain an invalid target, or a partition key that does not exist.
        # We will resolve these run requests, applying the target and partition config/tags.
        resolved_run_requests = []
        for run_request in run_requests:
            if run_request.job_name is None and has_multiple_targets:
                raise Exception(
                    f"Error in sensor {self._name}: Sensor returned a RunRequest that did not"
                    " specify job_name for the requested run. Expected one of:"
                    f" {target_names}"
                )
            elif (
                run_request.job_name
                and run_request.job_name not in target_names
                and not asset_selection
            ):
                raise Exception(
                    f"Error in sensor {self._name}: Sensor returned a RunRequest with job_name "
                    f"{run_request.job_name}. Expected one of: {target_names}"
                )

            if run_request.partition_key and not run_request.has_resolved_partition():
                selected_job = _get_repo_job_by_name(
                    context, run_request.job_name if run_request.job_name else target_names[0]
                )
                resolved_run_requests.append(
                    run_request.with_resolved_tags_and_config(
                        target_definition=selected_job,
                        current_time=None,
                        dynamic_partitions_store=dynamic_partitions_store,
                        dynamic_partitions_requests=dynamic_partitions_requests,
                    )
                )
            else:
                resolved_run_requests.append(run_request)

        return resolved_run_requests

    @property
    def _target(self) -> Optional[Union[DirectTarget, RepoRelativeTarget]]:
        return self._targets[0] if self._targets else None

    @public
    @property
    def job_name(self) -> Optional[str]:
        """Optional[str]: The name of the job that is targeted by this sensor."""
        if len(self._targets) > 1:
            raise DagsterInvalidInvocationError(
                f"Cannot use `job_name` property for sensor {self.name}, which targets multiple"
                " jobs."
            )
        return self._targets[0].job_name

    @public
    @property
    def default_status(self) -> DefaultSensorStatus:
        """DefaultSensorStatus: The default status for this sensor when it is first loaded in
        a code location.
        """
        return self._default_status

    @property
    def asset_selection(self) -> Optional[AssetSelection]:
        return self._asset_selection


@whitelist_for_serdes(
    storage_field_names={"dagster_run_reactions": "pipeline_run_reactions"},
)
class SensorExecutionData(
    NamedTuple(
        "_SensorExecutionData",
        [
            ("run_requests", Optional[Sequence[RunRequest]]),
            ("skip_message", Optional[str]),
            ("cursor", Optional[str]),
            ("dagster_run_reactions", Optional[Sequence[DagsterRunReaction]]),
            ("captured_log_key", Optional[Sequence[str]]),
            (
                "dynamic_partitions_requests",
                Optional[
                    Sequence[Union[AddDynamicPartitionsRequest, DeleteDynamicPartitionsRequest]]
                ],
            ),
            (
                "asset_events",
                Sequence[Union[AssetMaterialization, AssetObservation, AssetCheckEvaluation]],
            ),
        ],
    )
):
    dagster_run_reactions: Optional[Sequence[DagsterRunReaction]]

    def __new__(
        cls,
        run_requests: Optional[Sequence[RunRequest]] = None,
        skip_message: Optional[str] = None,
        cursor: Optional[str] = None,
        dagster_run_reactions: Optional[Sequence[DagsterRunReaction]] = None,
        captured_log_key: Optional[Sequence[str]] = None,
        dynamic_partitions_requests: Optional[
            Sequence[Union[AddDynamicPartitionsRequest, DeleteDynamicPartitionsRequest]]
        ] = None,
        asset_events: Optional[
            Sequence[Union[AssetMaterialization, AssetObservation, AssetCheckEvaluation]]
        ] = None,
    ):
        check.opt_sequence_param(run_requests, "run_requests", RunRequest)
        check.opt_str_param(skip_message, "skip_message")
        check.opt_str_param(cursor, "cursor")
        check.opt_sequence_param(dagster_run_reactions, "dagster_run_reactions", DagsterRunReaction)
        check.opt_list_param(captured_log_key, "captured_log_key", str)
        check.opt_sequence_param(
            dynamic_partitions_requests,
            "dynamic_partitions_requests",
            (AddDynamicPartitionsRequest, DeleteDynamicPartitionsRequest),
        )
        check.opt_sequence_param(
            asset_events,
            "asset_events",
            (AssetMaterialization, AssetObservation, AssetCheckEvaluation),
        )
        check.invariant(
            not (run_requests and skip_message), "Found both skip data and run request data"
        )
        return super(SensorExecutionData, cls).__new__(
            cls,
            run_requests=run_requests,
            skip_message=skip_message,
            cursor=cursor,
            dagster_run_reactions=dagster_run_reactions,
            captured_log_key=captured_log_key,
            dynamic_partitions_requests=dynamic_partitions_requests,
            asset_events=asset_events or [],
        )


def wrap_sensor_evaluation(
    sensor_name: str,
    fn: RawSensorEvaluationFunction,
) -> SensorEvaluationFunction:
    resource_arg_names: Set[str] = {arg.name for arg in get_resource_args(fn)}

    def _wrapped_fn(context: SensorEvaluationContext):
        resource_args_populated = validate_and_get_resource_dict(
            context.resources, sensor_name, resource_arg_names
        )

        context_param_name_if_present = get_context_param_name(fn)
        context_param = (
            {context_param_name_if_present: context} if context_param_name_if_present else {}
        )
        raw_evaluation_result = fn(**context_param, **resource_args_populated)

        def check_returned_scalar(scalar):
            if isinstance(scalar, (SkipReason, RunRequest, SensorResult)):
                return scalar
            elif scalar is not None:
                raise Exception(
                    f"Error in sensor {sensor_name}: Sensor unexpectedly returned output "
                    f"{scalar} of type {type(scalar)}.  Should only return SkipReason or "
                    "RunRequest objects."
                )

        if inspect.isgenerator(raw_evaluation_result):
            result = []
            try:
                while True:
                    result.append(next(raw_evaluation_result))
            except StopIteration as e:
                # captures the case where the evaluation function has a yield and also returns a
                # value
                if e.value is not None:
                    result.append(check_returned_scalar(e.value))

            return result
        elif isinstance(raw_evaluation_result, list):
            return raw_evaluation_result
        else:
            return [check_returned_scalar(raw_evaluation_result)]

    return _wrapped_fn


def build_sensor_context(
    instance: Optional[DagsterInstance] = None,
    cursor: Optional[str] = None,
    repository_name: Optional[str] = None,
    repository_def: Optional["RepositoryDefinition"] = None,
    sensor_name: Optional[str] = None,
    resources: Optional[Mapping[str, object]] = None,
    definitions: Optional["Definitions"] = None,
    instance_ref: Optional["InstanceRef"] = None,
) -> SensorEvaluationContext:
    """Builds sensor execution context using the provided parameters.

    This function can be used to provide a context to the invocation of a sensor definition.If
    provided, the dagster instance must be persistent; DagsterInstance.ephemeral() will result in an
    error.

    Args:
        instance (Optional[DagsterInstance]): The dagster instance configured to run the sensor.
        cursor (Optional[str]): A cursor value to provide to the evaluation of the sensor.
        repository_name (Optional[str]): The name of the repository that the sensor belongs to.
        repository_def (Optional[RepositoryDefinition]): The repository that the sensor belongs to.
            If needed by the sensor top-level resource definitions will be pulled from this repository.
            You can provide either this or `definitions`.
        resources (Optional[Mapping[str, ResourceDefinition]]): A set of resource definitions
            to provide to the sensor. If passed, these will override any resource definitions
            provided by the repository.
        definitions (Optional[Definitions]): `Definitions` object that the sensor is defined in.
            If needed by the sensor, top-level resource definitions will be pulled from these
            definitions. You can provide either this or `repository_def`.

    Examples:
        .. code-block:: python

            context = build_sensor_context()
            my_sensor(context)

    """
    from dagster._core.definitions.definitions_class import Definitions
    from dagster._core.definitions.repository_definition import RepositoryDefinition
    from dagster._core.execution.build_resources import wrap_resources_for_execution

    check.opt_inst_param(instance, "instance", DagsterInstance)
    check.opt_str_param(cursor, "cursor")
    check.opt_str_param(repository_name, "repository_name")
    repository_def = normalize_to_repository(
        check.opt_inst_param(definitions, "definitions", Definitions),
        check.opt_inst_param(repository_def, "repository_def", RepositoryDefinition),
        error_on_none=False,
    )

    return SensorEvaluationContext(
        instance_ref=instance_ref,
        last_completion_time=None,
        last_run_key=None,
        cursor=cursor,
        repository_name=repository_name,
        instance=instance,
        repository_def=repository_def,
        sensor_name=sensor_name,
        resources=wrap_resources_for_execution(resources),
    )


T = TypeVar("T")


def get_sensor_context_from_args_or_kwargs(
    fn: Callable,
    args: Tuple[Any, ...],
    kwargs: Dict[str, Any],
    context_type: Type[T],
) -> Optional[T]:
    from dagster._config.pythonic_config import is_coercible_to_resource

    context_param_name = get_context_param_name(fn)

    kwarg_keys_non_resource = set(kwargs.keys()) - {param.name for param in get_resource_args(fn)}
    if len(args) + len(kwarg_keys_non_resource) > 1:
        raise DagsterInvalidInvocationError(
            "Sensor invocation received multiple non-resource arguments. Only a first "
            "positional context parameter should be provided when invoking."
        )

    if any(is_coercible_to_resource(arg) for arg in args):
        raise DagsterInvalidInvocationError(
            "If directly invoking a sensor, you may not provide resources as"
            " positional"
            " arguments, only as keyword arguments."
        )

    context: Optional[T] = None

    if len(args) > 0:
        context = check.opt_inst(args[0], context_type)
    elif len(kwargs) > 0:
        if context_param_name and context_param_name not in kwargs:
            raise DagsterInvalidInvocationError(
                f"Sensor invocation expected argument '{context_param_name}'."
            )
        context = check.opt_inst(kwargs.get(context_param_name or "context"), context_type)
    elif context_param_name:
        # If the context parameter is present but no value was provided, we error
        raise DagsterInvalidInvocationError(
            "Sensor evaluation function expected context argument, but no context argument "
            "was provided when invoking."
        )

    return context


def get_or_create_sensor_context(
    fn: Callable,
    *args: Any,
    **kwargs: Any,
) -> SensorEvaluationContext:
    """Based on the passed resource function and the arguments passed to it, returns the
    user-passed SensorEvaluationContext or creates one if it is not passed.

    Raises an exception if the user passes more than one argument or if the user-provided
    function requires a context parameter but none is passed.
    """
    context = (
        get_sensor_context_from_args_or_kwargs(
            fn,
            args,
            kwargs,
            context_type=SensorEvaluationContext,
        )
        or build_sensor_context()
    )
    resource_args_from_kwargs = {}

    resource_args = {param.name for param in get_resource_args(fn)}
    for resource_arg in resource_args:
        if resource_arg in kwargs:
            resource_args_from_kwargs[resource_arg] = kwargs[resource_arg]

    if resource_args_from_kwargs:
        return context.merge_resources(resource_args_from_kwargs)

    return context


def _run_requests_with_base_asset_jobs(
    run_requests: Iterable[RunRequest],
    context: SensorEvaluationContext,
    outer_asset_selection: AssetSelection,
) -> Sequence[RunRequest]:
    """For sensors that target asset selections instead of jobs, finds the corresponding base asset
    for a selected set of assets.
    """
    asset_graph = context.repository_def.asset_graph  # type: ignore  # (possible none)
    result = []
    for run_request in run_requests:
        if run_request.asset_selection:
            asset_keys = run_request.asset_selection

            unexpected_asset_keys = (
                AssetSelection.keys(*asset_keys) - outer_asset_selection
            ).resolve(asset_graph)
            if unexpected_asset_keys:
                raise DagsterInvalidSubsetError(
                    "RunRequest includes asset keys that are not part of sensor's asset_selection:"
                    f" {unexpected_asset_keys}"
                )
        else:
            asset_keys = outer_asset_selection.resolve(asset_graph)

        base_job = context.repository_def.get_implicit_job_def_for_assets(asset_keys)  # type: ignore  # (possible none)
        result.append(
            run_request.with_replaced_attrs(
                job_name=base_job.name, asset_selection=list(asset_keys)  # type: ignore  # (possible none)
            )
        )

    return result



from datetime import datetime
from typing import Callable, Mapping, NamedTuple, Optional, Sequence, Union

from typing_extensions import TypeAlias

import dagster._check as check
from dagster._annotations import PublicAttr
from dagster._core.definitions.events import AssetKey, AssetMaterialization, AssetObservation
from dagster._core.errors import DagsterInvalidInvocationError
from dagster._core.events import DagsterEventType
from dagster._core.events.log import EventLogEntry
from dagster._serdes import whitelist_for_serdes

EventHandlerFn: TypeAlias = Callable[[EventLogEntry, str], None]


class RunShardedEventsCursor(NamedTuple):
    """Pairs an id-based event log cursor with a timestamp-based run cursor, for improved
    performance on run-sharded event log storages (e.g. the default SqliteEventLogStorage). For
    run-sharded storages, the id field is ignored, since they may not be unique across shards.
    """

    id: int
    run_updated_after: datetime


@whitelist_for_serdes
class EventLogRecord(NamedTuple):
    """Internal representation of an event record, as stored in a
    :py:class:`~dagster._core.storage.event_log.EventLogStorage`.

    Users should not instantiate this class directly.
    """

    storage_id: PublicAttr[int]
    event_log_entry: PublicAttr[EventLogEntry]

    @property
    def run_id(self) -> str:
        return self.event_log_entry.run_id

    @property
    def timestamp(self) -> float:
        return self.event_log_entry.timestamp

    @property
    def asset_key(self) -> Optional[AssetKey]:
        dagster_event = self.event_log_entry.dagster_event
        if dagster_event:
            return dagster_event.asset_key

        return None

    @property
    def partition_key(self) -> Optional[str]:
        dagster_event = self.event_log_entry.dagster_event
        if dagster_event:
            return dagster_event.partition

        return None

    @property
    def asset_materialization(self) -> Optional[AssetMaterialization]:
        return self.event_log_entry.asset_materialization

    @property
    def asset_observation(self) -> Optional[AssetObservation]:
        return self.event_log_entry.asset_observation


@whitelist_for_serdes
class EventRecordsFilter(
    NamedTuple(
        "_EventRecordsFilter",
        [
            ("event_type", DagsterEventType),
            ("asset_key", Optional[AssetKey]),
            ("asset_partitions", Optional[Sequence[str]]),
            ("after_cursor", Optional[Union[int, RunShardedEventsCursor]]),
            ("before_cursor", Optional[Union[int, RunShardedEventsCursor]]),
            ("after_timestamp", Optional[float]),
            ("before_timestamp", Optional[float]),
            ("storage_ids", Optional[Sequence[int]]),
            ("tags", Optional[Mapping[str, Union[str, Sequence[str]]]]),
        ],
    )
):
    """Defines a set of filter fields for fetching a set of event log entries or event log records.

    Args:
        event_type (DagsterEventType): Filter argument for dagster event type
        asset_key (Optional[AssetKey]): Asset key for which to get asset materialization event
            entries / records.
        asset_partitions (Optional[List[str]]): Filter parameter such that only asset
            events with a partition value matching one of the provided values.  Only
            valid when the `asset_key` parameter is provided.
        after_cursor (Optional[Union[int, RunShardedEventsCursor]]): Filter parameter such that only
            records with storage_id greater than the provided value are returned. Using a
            run-sharded events cursor will result in a significant performance gain when run against
            a SqliteEventLogStorage implementation (which is run-sharded)
        before_cursor (Optional[Union[int, RunShardedEventsCursor]]): Filter parameter such that
            records with storage_id less than the provided value are returned. Using a run-sharded
            events cursor will result in a significant performance gain when run against
            a SqliteEventLogStorage implementation (which is run-sharded)
        after_timestamp (Optional[float]): Filter parameter such that only event records for
            events with timestamp greater than the provided value are returned.
        before_timestamp (Optional[float]): Filter parameter such that only event records for
            events with timestamp less than the provided value are returned.
    """

    def __new__(
        cls,
        event_type: DagsterEventType,
        asset_key: Optional[AssetKey] = None,
        asset_partitions: Optional[Sequence[str]] = None,
        after_cursor: Optional[Union[int, RunShardedEventsCursor]] = None,
        before_cursor: Optional[Union[int, RunShardedEventsCursor]] = None,
        after_timestamp: Optional[float] = None,
        before_timestamp: Optional[float] = None,
        storage_ids: Optional[Sequence[int]] = None,
        tags: Optional[Mapping[str, Union[str, Sequence[str]]]] = None,
    ):
        check.opt_sequence_param(asset_partitions, "asset_partitions", of_type=str)
        check.inst_param(event_type, "event_type", DagsterEventType)

        tags = check.opt_mapping_param(tags, "tags", key_type=str)
        if tags and event_type is not DagsterEventType.ASSET_MATERIALIZATION:
            raise DagsterInvalidInvocationError(
                "Can only filter by tags for asset materialization events"
            )

        # type-ignores work around mypy type inference bug
        return super(EventRecordsFilter, cls).__new__(
            cls,
            event_type=event_type,
            asset_key=check.opt_inst_param(asset_key, "asset_key", AssetKey),
            asset_partitions=asset_partitions,
            after_cursor=check.opt_inst_param(
                after_cursor, "after_cursor", (int, RunShardedEventsCursor)
            ),
            before_cursor=check.opt_inst_param(
                before_cursor, "before_cursor", (int, RunShardedEventsCursor)
            ),
            after_timestamp=check.opt_float_param(after_timestamp, "after_timestamp"),
            before_timestamp=check.opt_float_param(before_timestamp, "before_timestamp"),
            storage_ids=check.opt_nullable_sequence_param(storage_ids, "storage_ids", of_type=int),
            tags=check.opt_mapping_param(tags, "tags", key_type=str),
        )



from typing import TYPE_CHECKING, Any, Mapping, Optional, Sequence

import dagster._check as check
from dagster._core.definitions.assets_job import build_assets_job
from dagster._core.definitions.definitions_class import Definitions
from dagster._utils.warnings import disable_dagster_warnings

from ..instance import DagsterInstance
from .source_asset import SourceAsset

if TYPE_CHECKING:
    from ..execution.execute_in_process_result import ExecuteInProcessResult


def observe(
    source_assets: Sequence[SourceAsset],
    run_config: Any = None,
    instance: Optional[DagsterInstance] = None,
    resources: Optional[Mapping[str, object]] = None,
    partition_key: Optional[str] = None,
    raise_on_error: bool = True,
    tags: Optional[Mapping[str, str]] = None,
) -> "ExecuteInProcessResult":
    """Executes a single-threaded, in-process run which observes provided source assets.

    By default, will materialize assets to the local filesystem.

    Args:
        source_assets (Sequence[SourceAsset]):
            The source assets to materialize.
        resources (Optional[Mapping[str, object]]):
            The resources needed for execution. Can provide resource instances
            directly, or resource definitions. Note that if provided resources
            conflict with resources directly on assets, an error will be thrown.
        run_config (Optional[Any]): The run config to use for the run that materializes the assets.
        partition_key: (Optional[str])
            The string partition key that specifies the run config to execute. Can only be used
            to select run config for assets with partitioned config.
        tags (Optional[Mapping[str, str]]): Tags for the run.

    Returns:
        ExecuteInProcessResult: The result of the execution.
    """
    source_assets = check.sequence_param(source_assets, "assets", of_type=(SourceAsset))
    instance = check.opt_inst_param(instance, "instance", DagsterInstance)
    partition_key = check.opt_str_param(partition_key, "partition_key")
    resources = check.opt_mapping_param(resources, "resources", key_type=str)

    with disable_dagster_warnings():
        observation_job = build_assets_job("in_process_observation_job", [], source_assets)
        defs = Definitions(
            assets=source_assets,
            jobs=[observation_job],
            resources=resources,
        )

        return defs.get_job_def("in_process_observation_job").execute_in_process(
            run_config=run_config,
            instance=instance,
            partition_key=partition_key,
            raise_on_error=raise_on_error,
            tags=tags,
        )



import dagster._check as check
from dagster._config import ConfigSchemaSnapshot, snap_from_config_type
from dagster._core.definitions.job_definition import JobDefinition


def build_config_schema_snapshot(job_def: JobDefinition) -> ConfigSchemaSnapshot:
    check.inst_param(job_def, "job_def", JobDefinition)
    config_snaps_by_key = {
        ct.key: snap_from_config_type(ct) for ct in job_def.run_config_schema.all_config_types()
    }
    return ConfigSchemaSnapshot(config_snaps_by_key)



import warnings
from contextlib import contextmanager
from typing import Callable, Iterator, Optional, TypeVar

import dagster._check as check
from dagster._core.decorator_utils import (
    Decoratable,
    apply_context_manager_decorator,
)

T = TypeVar("T")

# ########################
# ##### DEPRECATED
# ########################


def normalize_renamed_param(
    new_val: T,
    new_arg: str,
    old_val: T,
    old_arg: str,
    coerce_old_to_new: Optional[Callable[[T], T]] = None,
) -> T:
    """Utility for managing backwards compatibility of a renamed parameter.

    .. code-block::

       # The name of param `old_flag` is being updated to `new_flag`, but we are temporarily
       # accepting either param.
       def is_new(old_flag=None, new_flag=None):
           return canonicalize_backcompat_args(
               new_val=new_flag,
               new_arg='new_flag',
               old_val=old_flag,
               old_arg='old_flag',
               breaking_version='0.9.0',
               coerce_old_to_new=lambda val: not val,
           )

    In the above example, if the caller sets both new_flag and old_flag, it will fail by throwing
    a CheckError. If the caller sets the new_flag, it's returned unaltered. If the caller sets
    old_flag, it will return the old_flag run through the coercion function.
    """
    check.str_param(new_arg, "new_arg")
    check.str_param(old_arg, "old_arg")
    check.opt_callable_param(coerce_old_to_new, "coerce_old_to_new")
    if new_val is not None and old_val is not None:
        check.failed(f'Do not use deprecated "{old_arg}" now that you are using "{new_arg}".')
    elif old_val is not None:
        return coerce_old_to_new(old_val) if coerce_old_to_new else old_val
    else:
        return new_val


def deprecation_warning(
    subject: str,
    breaking_version: str,
    additional_warn_text: Optional[str] = None,
    stacklevel: int = 3,
):
    warnings.warn(
        f"{subject} is deprecated and will be removed in {breaking_version}."
        + ((" " + additional_warn_text) if additional_warn_text else ""),
        category=DeprecationWarning,
        stacklevel=stacklevel,
    )


# ########################
# ##### EXPERIMENTAL
# ########################

EXPERIMENTAL_WARNING_HELP = (
    "To mute warnings for experimental functionality, invoke"
    ' warnings.filterwarnings("ignore", category=dagster.ExperimentalWarning) or use'
    " one of the other methods described at"
    " https://docs.python.org/3/library/warnings.html#describing-warning-filters."
)


class ExperimentalWarning(Warning):
    pass


def experimental_warning(
    subject: str, additional_warn_text: Optional[str] = None, stacklevel: int = 3
) -> None:
    extra_text = f" {additional_warn_text}" if additional_warn_text else ""
    warnings.warn(
        f"{subject} is experimental. It may break in future versions, even between dot"
        f" releases.{extra_text} {EXPERIMENTAL_WARNING_HELP}",
        ExperimentalWarning,
        stacklevel=stacklevel,
    )


# ########################
# ##### DISABLE DAGSTER WARNINGS
# ########################


@contextmanager
def disable_dagster_warnings() -> Iterator[None]:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=DeprecationWarning)
        warnings.simplefilter("ignore", category=ExperimentalWarning)
        yield


T_Decoratable = TypeVar("T_Decoratable", bound=Decoratable)


def suppress_dagster_warnings(__obj: T_Decoratable) -> T_Decoratable:
    """Mark a method/function as ignoring Dagster-generated warnings. This suppresses any
    `ExperimentalWarnings` or `DeprecationWarnings` when the function is called.

    Usage:

        .. code-block:: python

            @suppress_dagster_warnings
            def invokes_some_experimental_stuff(my_arg):
                my_experimental_function(my_arg)
    """
    return apply_context_manager_decorator(__obj, disable_dagster_warnings)



from abc import ABC, ABCMeta, abstractmethod
from inspect import _empty as EmptyAnnotation
from typing import (
    AbstractSet,
    Any,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Union,
    cast,
)

import dagster._check as check
from dagster._annotations import deprecated, experimental, public
from dagster._core.definitions.asset_check_spec import AssetCheckKey, AssetCheckSpec
from dagster._core.definitions.asset_checks import AssetChecksDefinition
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.data_version import (
    DataProvenance,
    DataVersion,
    extract_data_provenance_from_entry,
)
from dagster._core.definitions.decorators.op_decorator import DecoratedOpFunction
from dagster._core.definitions.dependency import Node, NodeHandle
from dagster._core.definitions.events import (
    AssetKey,
    AssetMaterialization,
    AssetObservation,
    ExpectationResult,
    UserEvent,
)
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.partition_key_range import PartitionKeyRange
from dagster._core.definitions.step_launcher import StepLauncher
from dagster._core.definitions.time_window_partitions import TimeWindow
from dagster._core.errors import (
    DagsterInvalidDefinitionError,
    DagsterInvalidPropertyError,
    DagsterInvariantViolationError,
)
from dagster._core.events import DagsterEvent
from dagster._core.instance import DagsterInstance
from dagster._core.log_manager import DagsterLogManager
from dagster._core.storage.dagster_run import DagsterRun
from dagster._utils.forked_pdb import ForkedPdb
from dagster._utils.warnings import (
    deprecation_warning,
)

from .system import StepExecutionContext


# This metaclass has to exist for OpExecutionContext to have a metaclass
class AbstractComputeMetaclass(ABCMeta):
    pass


class AbstractComputeExecutionContext(ABC, metaclass=AbstractComputeMetaclass):
    """Base class for op context implemented by OpExecutionContext and DagstermillExecutionContext."""

    @abstractmethod
    def has_tag(self, key: str) -> bool:
        """Implement this method to check if a logging tag is set."""

    @abstractmethod
    def get_tag(self, key: str) -> Optional[str]:
        """Implement this method to get a logging tag."""

    @property
    @abstractmethod
    def run_id(self) -> str:
        """The run id for the context."""

    @property
    @abstractmethod
    def op_def(self) -> OpDefinition:
        """The op definition corresponding to the execution step being executed."""

    @property
    @abstractmethod
    def job_def(self) -> JobDefinition:
        """The job being executed."""

    @property
    @abstractmethod
    def run(self) -> DagsterRun:
        """The DagsterRun object corresponding to the execution."""

    @property
    @abstractmethod
    def resources(self) -> Any:
        """Resources available in the execution context."""

    @property
    @abstractmethod
    def log(self) -> DagsterLogManager:
        """The log manager available in the execution context."""

    @property
    @abstractmethod
    def op_config(self) -> Any:
        """The parsed config specific to this op."""


class OpExecutionContextMetaClass(AbstractComputeMetaclass):
    def __instancecheck__(cls, instance) -> bool:
        # This makes isinstance(context, OpExecutionContext) throw a deprecation warning when
        # context is an AssetExecutionContext. This metaclass can be deleted once AssetExecutionContext
        # has been split into it's own class in 1.7.0
        if type(instance) is AssetExecutionContext and cls is not AssetExecutionContext:
            deprecation_warning(
                subject="AssetExecutionContext",
                additional_warn_text=(
                    "Starting in version 1.7.0 AssetExecutionContext will no longer be a subclass"
                    " of OpExecutionContext."
                ),
                breaking_version="1.7.0",
                stacklevel=1,
            )
        return super().__instancecheck__(instance)


class OpExecutionContext(AbstractComputeExecutionContext, metaclass=OpExecutionContextMetaClass):
    """The ``context`` object that can be made available as the first argument to the function
    used for computing an op or asset.

    This context object provides system information such as resources, config, and logging.

    To construct an execution context for testing purposes, use :py:func:`dagster.build_op_context`.

    Example:
        .. code-block:: python

            from dagster import op, OpExecutionContext

            @op
            def hello_world(context: OpExecutionContext):
                context.log.info("Hello, world!")
    """

    __slots__ = ["_step_execution_context"]

    def __init__(self, step_execution_context: StepExecutionContext):
        self._step_execution_context = check.inst_param(
            step_execution_context,
            "step_execution_context",
            StepExecutionContext,
        )
        self._pdb: Optional[ForkedPdb] = None
        self._events: List[DagsterEvent] = []
        self._output_metadata: Dict[str, Any] = {}

    @public
    @property
    def op_config(self) -> Any:
        """Any: The parsed config specific to this op."""
        return self._step_execution_context.op_config

    @property
    def dagster_run(self) -> DagsterRun:
        """PipelineRun: The current pipeline run."""
        return self._step_execution_context.dagster_run

    @property
    def run(self) -> DagsterRun:
        """DagsterRun: The current run."""
        return self.dagster_run

    @public
    @property
    def instance(self) -> DagsterInstance:
        """DagsterInstance: The current Dagster instance."""
        return self._step_execution_context.instance

    @public
    @property
    def pdb(self) -> ForkedPdb:
        """dagster.utils.forked_pdb.ForkedPdb: Gives access to pdb debugging from within the op.

        Example:
            .. code-block:: python

                @op
                def debug(context):
                    context.pdb.set_trace()
        """
        if self._pdb is None:
            self._pdb = ForkedPdb()

        return self._pdb

    @property
    def file_manager(self):
        """Deprecated access to the file manager.

        :meta private:
        """
        raise DagsterInvalidPropertyError(
            "You have attempted to access the file manager which has been moved to resources in"
            " 0.10.0. Please access it via `context.resources.file_manager` instead."
        )

    @public
    @property
    def resources(self) -> Any:
        """Resources: The currently available resources."""
        return self._step_execution_context.resources

    @property
    def step_launcher(self) -> Optional[StepLauncher]:
        """Optional[StepLauncher]: The current step launcher, if any."""
        return self._step_execution_context.step_launcher

    @public
    @property
    def run_id(self) -> str:
        """str: The id of the current execution's run."""
        return self._step_execution_context.run_id

    @public
    @property
    def run_config(self) -> Mapping[str, object]:
        """dict: The run config for the current execution."""
        return self._step_execution_context.run_config

    @public
    @property
    def job_def(self) -> JobDefinition:
        """JobDefinition: The currently executing pipeline."""
        return self._step_execution_context.job_def

    @public
    @property
    def job_name(self) -> str:
        """str: The name of the currently executing pipeline."""
        return self._step_execution_context.job_name

    @public
    @property
    def log(self) -> DagsterLogManager:
        """DagsterLogManager: The log manager available in the execution context."""
        return self._step_execution_context.log

    @property
    def node_handle(self) -> NodeHandle:
        """NodeHandle: The current op's handle.

        :meta private:
        """
        return self._step_execution_context.node_handle

    @property
    def op_handle(self) -> NodeHandle:
        """NodeHandle: The current op's handle.

        :meta private:
        """
        return self.node_handle

    @property
    def op(self) -> Node:
        """Node: The object representing the invoked op within the graph.

        :meta private:

        """
        return self._step_execution_context.job_def.get_node(self.node_handle)

    @public
    @property
    def op_def(self) -> OpDefinition:
        """OpDefinition: The current op definition."""
        return cast(OpDefinition, self.op.definition)

    @public
    @property
    def has_partition_key(self) -> bool:
        """Whether the current run is a partitioned run."""
        return self._step_execution_context.has_partition_key

    @public
    @property
    def partition_key(self) -> str:
        """The partition key for the current run.

        Raises an error if the current run is not a partitioned run. Or if the current run is operating
        over a range of partitions (ie. a backfill of several partitions executed in a single run).

        Examples:
            .. code-block:: python

                partitions_def = DailyPartitionsDefinition("2023-08-20")

                @asset(
                    partitions_def=partitions_def
                )
                def my_asset(context: AssetExecutionContext):
                    context.log.info(context.partition_key)

                # materializing the 2023-08-21 partition of this asset will log:
                #   "2023-08-21"
        """
        return self._step_execution_context.partition_key

    @deprecated(breaking_version="2.0", additional_warn_text="Use `partition_key_range` instead.")
    @public
    @property
    def asset_partition_key_range(self) -> PartitionKeyRange:
        """The range of partition keys for the current run.

        If run is for a single partition key, return a `PartitionKeyRange` with the same start and
        end. Raises an error if the current run is not a partitioned run.
        """
        return self.partition_key_range

    @public
    @property
    def partition_key_range(self) -> PartitionKeyRange:
        """The range of partition keys for the current run.

        If run is for a single partition key, returns a `PartitionKeyRange` with the same start and
        end. Raises an error if the current run is not a partitioned run.

        Examples:
            .. code-block:: python

                partitions_def = DailyPartitionsDefinition("2023-08-20")

                @asset(
                    partitions_def=partitions_def
                )
                def my_asset(context: AssetExecutionContext):
                    context.log.info(context.partition_key_range)

                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   PartitionKeyRange(start="2023-08-21", end="2023-08-25")
        """
        return self._step_execution_context.asset_partition_key_range

    @public
    @property
    def partition_time_window(self) -> TimeWindow:
        """The partition time window for the current run.

        Raises an error if the current run is not a partitioned run, or if the job's partition
        definition is not a TimeWindowPartitionsDefinition.

        Examples:
            .. code-block:: python

                partitions_def = DailyPartitionsDefinition("2023-08-20")

                @asset(
                    partitions_def=partitions_def
                )
                def my_asset(context: AssetExecutionContext):
                    context.log.info(context.partition_time_window)

                # materializing the 2023-08-21 partition of this asset will log:
                #   TimeWindow("2023-08-21", "2023-08-22")
        """
        return self._step_execution_context.partition_time_window

    @public
    def has_tag(self, key: str) -> bool:
        """Check if a logging tag is set.

        Args:
            key (str): The tag to check.

        Returns:
            bool: Whether the tag is set.
        """
        return self._step_execution_context.has_tag(key)

    @public
    def get_tag(self, key: str) -> Optional[str]:
        """Get a logging tag.

        Args:
            key (tag): The tag to get.

        Returns:
            Optional[str]: The value of the tag, if present.
        """
        return self._step_execution_context.get_tag(key)

    @property
    def run_tags(self) -> Mapping[str, str]:
        """Mapping[str, str]: The tags for the current run."""
        return self._step_execution_context.run_tags

    def has_events(self) -> bool:
        return bool(self._events)

    def consume_events(self) -> Iterator[DagsterEvent]:
        """Pops and yields all user-generated events that have been recorded from this context.

        If consume_events has not yet been called, this will yield all logged events since the beginning of the op's computation. If consume_events has been called, it will yield all events since the last time consume_events was called. Designed for internal use. Users should never need to invoke this method.
        """
        events = self._events
        self._events = []
        yield from events

    @public
    def log_event(self, event: UserEvent) -> None:
        """Log an AssetMaterialization, AssetObservation, or ExpectationResult from within the body of an op.

        Events logged with this method will appear in the list of DagsterEvents, as well as the event log.

        Args:
            event (Union[AssetMaterialization, AssetObservation, ExpectationResult]): The event to log.

        **Examples:**

        .. code-block:: python

            from dagster import op, AssetMaterialization

            @op
            def log_materialization(context):
                context.log_event(AssetMaterialization("foo"))
        """
        if isinstance(event, AssetMaterialization):
            self._events.append(
                DagsterEvent.asset_materialization(self._step_execution_context, event)
            )
        elif isinstance(event, AssetObservation):
            self._events.append(DagsterEvent.asset_observation(self._step_execution_context, event))
        elif isinstance(event, ExpectationResult):
            self._events.append(
                DagsterEvent.step_expectation_result(self._step_execution_context, event)
            )
        else:
            check.failed(f"Unexpected event {event}")

    @public
    def add_output_metadata(
        self,
        metadata: Mapping[str, Any],
        output_name: Optional[str] = None,
        mapping_key: Optional[str] = None,
    ) -> None:
        """Add metadata to one of the outputs of an op.

        This can be invoked multiple times per output in the body of an op. If the same key is
        passed multiple times, the value associated with the last call will be used.

        Args:
            metadata (Mapping[str, Any]): The metadata to attach to the output
            output_name (Optional[str]): The name of the output to attach metadata to. If there is only one output on the op, then this argument does not need to be provided. The metadata will automatically be attached to the only output.
            mapping_key (Optional[str]): The mapping key of the output to attach metadata to. If the
                output is not dynamic, this argument does not need to be provided.

        **Examples:**

        .. code-block:: python

            from dagster import Out, op
            from typing import Tuple

            @op
            def add_metadata(context):
                context.add_output_metadata({"foo", "bar"})
                return 5 # Since the default output is called "result", metadata will be attached to the output "result".

            @op(out={"a": Out(), "b": Out()})
            def add_metadata_two_outputs(context) -> Tuple[str, int]:
                context.add_output_metadata({"foo": "bar"}, output_name="b")
                context.add_output_metadata({"baz": "bat"}, output_name="a")

                return ("dog", 5)

        """
        metadata = check.mapping_param(metadata, "metadata", key_type=str)
        output_name = check.opt_str_param(output_name, "output_name")
        mapping_key = check.opt_str_param(mapping_key, "mapping_key")

        self._step_execution_context.add_output_metadata(
            metadata=metadata, output_name=output_name, mapping_key=mapping_key
        )

    def get_output_metadata(
        self, output_name: str, mapping_key: Optional[str] = None
    ) -> Optional[Mapping[str, Any]]:
        return self._step_execution_context.get_output_metadata(
            output_name=output_name, mapping_key=mapping_key
        )

    def get_step_execution_context(self) -> StepExecutionContext:
        """Allows advanced users (e.g. framework authors) to punch through to the underlying
        step execution context.

        :meta private:

        Returns:
            StepExecutionContext: The underlying system context.
        """
        return self._step_execution_context

    @public
    @property
    def retry_number(self) -> int:
        """Which retry attempt is currently executing i.e. 0 for initial attempt, 1 for first retry, etc."""
        return self._step_execution_context.previous_attempt_count

    def describe_op(self):
        return self._step_execution_context.describe_op()

    @public
    def get_mapping_key(self) -> Optional[str]:
        """Which mapping_key this execution is for if downstream of a DynamicOutput, otherwise None."""
        return self._step_execution_context.step.get_mapping_key()

    #############################################################################################
    # asset related methods
    #############################################################################################

    @public
    @property
    def asset_key(self) -> AssetKey:
        """The AssetKey for the current asset. In a multi_asset, use asset_key_for_output instead."""
        if self.has_assets_def and len(self.assets_def.keys_by_output_name.keys()) > 1:
            raise DagsterInvariantViolationError(
                "Cannot call `context.asset_key` in a multi_asset with more than one asset. Use"
                " `context.asset_key_for_output` instead."
            )
        return next(iter(self.assets_def.keys_by_output_name.values()))

    @public
    @property
    def has_assets_def(self) -> bool:
        """If there is a backing AssetsDefinition for what is currently executing."""
        assets_def = self.job_def.asset_layer.assets_def_for_node(self.node_handle)
        return assets_def is not None

    @public
    @property
    def assets_def(self) -> AssetsDefinition:
        """The backing AssetsDefinition for what is currently executing, errors if not available."""
        assets_def = self.job_def.asset_layer.assets_def_for_node(self.node_handle)
        if assets_def is None:
            raise DagsterInvalidPropertyError(
                f"Op '{self.op.name}' does not have an assets definition."
            )
        return assets_def

    @public
    @property
    def selected_asset_keys(self) -> AbstractSet[AssetKey]:
        """Get the set of AssetKeys this execution is expected to materialize."""
        if not self.has_assets_def:
            return set()
        return self.assets_def.keys

    @public
    @property
    def has_asset_checks_def(self) -> bool:
        """Return a boolean indicating the presence of a backing AssetChecksDefinition
        for the current execution.

        Returns:
            bool: True if there is a backing AssetChecksDefinition for the current execution, otherwise False.
        """
        return self.job_def.asset_layer.asset_checks_def_for_node(self.node_handle) is not None

    @public
    @property
    def asset_checks_def(self) -> AssetChecksDefinition:
        """The backing AssetChecksDefinition for what is currently executing, errors if not
        available.

        Returns:
            AssetChecksDefinition.
        """
        asset_checks_def = self.job_def.asset_layer.asset_checks_def_for_node(self.node_handle)
        if asset_checks_def is None:
            raise DagsterInvalidPropertyError(
                f"Op '{self.op.name}' does not have an asset checks definition."
            )

        return asset_checks_def

    @public
    @property
    def selected_asset_check_keys(self) -> AbstractSet[AssetCheckKey]:
        if self.has_assets_def:
            return self.assets_def.check_keys

        if self.has_asset_checks_def:
            check.failed("Subset selection is not yet supported within an AssetChecksDefinition")

        return set()

    @public
    @property
    def selected_output_names(self) -> AbstractSet[str]:
        """Get the output names that correspond to the current selection of assets this execution is expected to materialize."""
        # map selected asset keys to the output names they correspond to
        selected_asset_keys = self.selected_asset_keys
        selected_outputs: Set[str] = set()
        for output_name in self.op.output_dict.keys():
            asset_info = self.job_def.asset_layer.asset_info_for_output(
                self.node_handle, output_name
            )
            if any(  #  For graph-backed assets, check if a downstream asset is selected
                [
                    asset_key in selected_asset_keys
                    for asset_key in self.job_def.asset_layer.downstream_dep_assets(
                        self.node_handle, output_name
                    )
                ]
            ) or (asset_info and asset_info.key in selected_asset_keys):
                selected_outputs.add(output_name)

        return selected_outputs

    @public
    def asset_key_for_output(self, output_name: str = "result") -> AssetKey:
        """Return the AssetKey for the corresponding output."""
        asset_output_info = self.job_def.asset_layer.asset_info_for_output(
            node_handle=self.op_handle, output_name=output_name
        )
        if asset_output_info is None:
            check.failed(f"Output '{output_name}' has no asset")
        else:
            return asset_output_info.key

    @public
    def output_for_asset_key(self, asset_key: AssetKey) -> str:
        """Return the output name for the corresponding asset key."""
        node_output_handle = self.job_def.asset_layer.node_output_handle_for_asset(asset_key)
        if node_output_handle is None:
            check.failed(f"Asset key '{asset_key}' has no output")
        else:
            return node_output_handle.output_name

    @public
    def asset_key_for_input(self, input_name: str) -> AssetKey:
        """Return the AssetKey for the corresponding input."""
        key = self.job_def.asset_layer.asset_key_for_input(
            node_handle=self.op_handle, input_name=input_name
        )
        if key is None:
            check.failed(f"Input '{input_name}' has no asset")
        else:
            return key

    @public
    def asset_partition_key_for_output(self, output_name: str = "result") -> str:
        """Returns the asset partition key for the given output.

        Args:
            output_name (str): For assets defined with the ``@asset`` decorator, the name of the output
                will be automatically provided. For assets defined with ``@multi_asset``, ``output_name``
                should be the op output associated with the asset key (as determined by AssetOut)
                to get the partition key for.

        Examples:
            .. code-block:: python

                partitions_def = DailyPartitionsDefinition("2023-08-20")

                @asset(
                    partitions_def=partitions_def
                )
                def an_asset(context: AssetExecutionContext):
                    context.log.info(context.asset_partition_key_for_output())


                # materializing the 2023-08-21 partition of this asset will log:
                #   "2023-08-21"

                @multi_asset(
                    outs={
                        "first_asset": AssetOut(key=["my_assets", "first_asset"]),
                        "second_asset": AssetOut(key=["my_assets", "second_asset"])
                    }
                    partitions_def=partitions_def,
                )
                def a_multi_asset(context: AssetExecutionContext):
                    context.log.info(context.asset_partition_key_for_output("first_asset"))
                    context.log.info(context.asset_partition_key_for_output("second_asset"))


                # materializing the 2023-08-21 partition of this asset will log:
                #   "2023-08-21"
                #   "2023-08-21"


                @asset(
                    partitions_def=partitions_def,
                    ins={
                        "self_dependent_asset": AssetIn(partition_mapping=TimeWindowPartitionMapping(start_offset=-1, end_offset=-1))
                    }
                )
                def self_dependent_asset(context: AssetExecutionContext, self_dependent_asset):
                    context.log.info(context.asset_partition_key_for_output())

                # materializing the 2023-08-21 partition of this asset will log:
                #   "2023-08-21"

        """
        return self._step_execution_context.asset_partition_key_for_output(output_name)

    @public
    def asset_partitions_time_window_for_output(self, output_name: str = "result") -> TimeWindow:
        """The time window for the partitions of the output asset.

        If you want to write your asset to support running a backfill of several partitions in a single run,
        you can use ``asset_partitions_time_window_for_output`` to get the TimeWindow of all of the partitions
        being materialized by the backfill.

        Raises an error if either of the following are true:
        - The output asset has no partitioning.
        - The output asset is not partitioned with a TimeWindowPartitionsDefinition or a
        MultiPartitionsDefinition with one time-partitioned dimension.

        Args:
            output_name (str): For assets defined with the ``@asset`` decorator, the name of the output
                will be automatically provided. For assets defined with ``@multi_asset``, ``output_name``
                should be the op output associated with the asset key (as determined by AssetOut)
                to get the time window for.

        Examples:
            .. code-block:: python

                partitions_def = DailyPartitionsDefinition("2023-08-20")

                @asset(
                    partitions_def=partitions_def
                )
                def an_asset(context: AssetExecutionContext):
                    context.log.info(context.asset_partitions_time_window_for_output())


                # materializing the 2023-08-21 partition of this asset will log:
                #   TimeWindow("2023-08-21", "2023-08-22")

                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   TimeWindow("2023-08-21", "2023-08-26")

                @multi_asset(
                    outs={
                        "first_asset": AssetOut(key=["my_assets", "first_asset"]),
                        "second_asset": AssetOut(key=["my_assets", "second_asset"])
                    }
                    partitions_def=partitions_def,
                )
                def a_multi_asset(context: AssetExecutionContext):
                    context.log.info(context.asset_partitions_time_window_for_output("first_asset"))
                    context.log.info(context.asset_partitions_time_window_for_output("second_asset"))

                # materializing the 2023-08-21 partition of this asset will log:
                #   TimeWindow("2023-08-21", "2023-08-22")
                #   TimeWindow("2023-08-21", "2023-08-22")

                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   TimeWindow("2023-08-21", "2023-08-26")
                #   TimeWindow("2023-08-21", "2023-08-26")


                @asset(
                    partitions_def=partitions_def,
                    ins={
                        "self_dependent_asset": AssetIn(partition_mapping=TimeWindowPartitionMapping(start_offset=-1, end_offset=-1))
                    }
                )
                def self_dependent_asset(context: AssetExecutionContext, self_dependent_asset):
                    context.log.info(context.asset_partitions_time_window_for_output())

                # materializing the 2023-08-21 partition of this asset will log:
                #   TimeWindow("2023-08-21", "2023-08-22")

                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   TimeWindow("2023-08-21", "2023-08-26")

        """
        return self._step_execution_context.asset_partitions_time_window_for_output(output_name)

    @public
    def asset_partition_key_range_for_output(
        self, output_name: str = "result"
    ) -> PartitionKeyRange:
        """Return the PartitionKeyRange for the corresponding output. Errors if the run is not partitioned.

        If you want to write your asset to support running a backfill of several partitions in a single run,
        you can use ``asset_partition_key_range_for_output`` to get all of the partitions being materialized
        by the backfill.

        Args:
            output_name (str): For assets defined with the ``@asset`` decorator, the name of the output
                will be automatically provided. For assets defined with ``@multi_asset``, ``output_name``
                should be the op output associated with the asset key (as determined by AssetOut)
                to get the partition key range for.

        Examples:
            .. code-block:: python

                partitions_def = DailyPartitionsDefinition("2023-08-20")

                @asset(
                    partitions_def=partitions_def
                )
                def an_asset(context: AssetExecutionContext):
                    context.log.info(context.asset_partition_key_range_for_output())


                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   PartitionKeyRange(start="2023-08-21", end="2023-08-25")

                @multi_asset(
                    outs={
                        "first_asset": AssetOut(key=["my_assets", "first_asset"]),
                        "second_asset": AssetOut(key=["my_assets", "second_asset"])
                    }
                    partitions_def=partitions_def,
                )
                def a_multi_asset(context: AssetExecutionContext):
                    context.log.info(context.asset_partition_key_range_for_output("first_asset"))
                    context.log.info(context.asset_partition_key_range_for_output("second_asset"))


                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   PartitionKeyRange(start="2023-08-21", end="2023-08-25")
                #   PartitionKeyRange(start="2023-08-21", end="2023-08-25")


                @asset(
                    partitions_def=partitions_def,
                    ins={
                        "self_dependent_asset": AssetIn(partition_mapping=TimeWindowPartitionMapping(start_offset=-1, end_offset=-1))
                    }
                )
                def self_dependent_asset(context: AssetExecutionContext, self_dependent_asset):
                    context.log.info(context.asset_partition_key_range_for_output())

                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   PartitionKeyRange(start="2023-08-21", end="2023-08-25")

        """
        return self._step_execution_context.asset_partition_key_range_for_output(output_name)

    @public
    def asset_partition_key_range_for_input(self, input_name: str) -> PartitionKeyRange:
        """Return the PartitionKeyRange for the corresponding input. Errors if the asset depends on a
        non-contiguous chunk of the input.

        If you want to write your asset to support running a backfill of several partitions in a single run,
        you can use ``asset_partition_key_range_for_input`` to get the range of partitions keys of the input that
        are relevant to that backfill.

        Args:
            input_name (str): The name of the input to get the time window for.

        Examples:
            .. code-block:: python

                partitions_def = DailyPartitionsDefinition("2023-08-20")

                @asset(
                    partitions_def=partitions_def
                )
                def upstream_asset():
                    ...

                @asset(
                    partitions_def=partitions_def
                )
                def an_asset(context: AssetExecutionContext, upstream_asset):
                    context.log.info(context.asset_partition_key_range_for_input("upstream_asset"))


                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   PartitionKeyRange(start="2023-08-21", end="2023-08-25")

                @asset(
                    ins={
                        "upstream_asset": AssetIn(partition_mapping=TimeWindowPartitionMapping(start_offset=-1, end_offset=-1))
                    }
                    partitions_def=partitions_def,
                )
                def another_asset(context: AssetExecutionContext, upstream_asset):
                    context.log.info(context.asset_partition_key_range_for_input("upstream_asset"))


                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   PartitionKeyRange(start="2023-08-20", end="2023-08-24")


                @asset(
                    partitions_def=partitions_def,
                    ins={
                        "self_dependent_asset": AssetIn(partition_mapping=TimeWindowPartitionMapping(start_offset=-1, end_offset=-1))
                    }
                )
                def self_dependent_asset(context: AssetExecutionContext, self_dependent_asset):
                    context.log.info(context.asset_partition_key_range_for_input("self_dependent_asset"))

                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   PartitionKeyRange(start="2023-08-20", end="2023-08-24")


        """
        return self._step_execution_context.asset_partition_key_range_for_input(input_name)

    @public
    def asset_partition_key_for_input(self, input_name: str) -> str:
        """Returns the partition key of the upstream asset corresponding to the given input.

        Args:
            input_name (str): The name of the input to get the partition key for.

        Examples:
            .. code-block:: python

                partitions_def = DailyPartitionsDefinition("2023-08-20")

                @asset(
                    partitions_def=partitions_def
                )
                def upstream_asset():
                    ...

                @asset(
                    partitions_def=partitions_def
                )
                def an_asset(context: AssetExecutionContext, upstream_asset):
                    context.log.info(context.asset_partition_key_for_input("upstream_asset"))

                # materializing the 2023-08-21 partition of this asset will log:
                #   "2023-08-21"


                @asset(
                    partitions_def=partitions_def,
                    ins={
                        "self_dependent_asset": AssetIn(partition_mapping=TimeWindowPartitionMapping(start_offset=-1, end_offset=-1))
                    }
                )
                def self_dependent_asset(context: AssetExecutionContext, self_dependent_asset):
                    context.log.info(context.asset_partition_key_for_input("self_dependent_asset"))

                # materializing the 2023-08-21 partition of this asset will log:
                #   "2023-08-20"

        """
        return self._step_execution_context.asset_partition_key_for_input(input_name)

    @public
    def asset_partitions_def_for_output(self, output_name: str = "result") -> PartitionsDefinition:
        """The PartitionsDefinition on the asset corresponding to this output.

        Args:
            output_name (str): For assets defined with the ``@asset`` decorator, the name of the output
                will be automatically provided. For assets defined with ``@multi_asset``, ``output_name``
                should be the op output associated with the asset key (as determined by AssetOut)
                to get the PartitionsDefinition for.

        Examples:
            .. code-block:: python

                partitions_def = DailyPartitionsDefinition("2023-08-20")

                @asset(
                    partitions_def=partitions_def
                )
                def upstream_asset(context: AssetExecutionContext):
                    context.log.info(context.asset_partitions_def_for_output())

                # materializing the 2023-08-21 partition of this asset will log:
                #   DailyPartitionsDefinition("2023-08-20")

                @multi_asset(
                    outs={
                        "first_asset": AssetOut(key=["my_assets", "first_asset"]),
                        "second_asset": AssetOut(key=["my_assets", "second_asset"])
                    }
                    partitions_def=partitions_def,
                )
                def a_multi_asset(context: AssetExecutionContext):
                    context.log.info(context.asset_partitions_def_for_output("first_asset"))
                    context.log.info(context.asset_partitions_def_for_output("second_asset"))

                # materializing the 2023-08-21 partition of this asset will log:
                #   DailyPartitionsDefinition("2023-08-20")
                #   DailyPartitionsDefinition("2023-08-20")

        """
        asset_key = self.asset_key_for_output(output_name)
        result = self._step_execution_context.job_def.asset_layer.partitions_def_for_asset(
            asset_key
        )
        if result is None:
            raise DagsterInvariantViolationError(
                f"Attempting to access partitions def for asset {asset_key}, but it is not"
                " partitioned"
            )

        return result

    @public
    def asset_partitions_def_for_input(self, input_name: str) -> PartitionsDefinition:
        """The PartitionsDefinition on the upstream asset corresponding to this input.

        Args:
            input_name (str): The name of the input to get the PartitionsDefinition for.

        Examples:
            .. code-block:: python

                partitions_def = DailyPartitionsDefinition("2023-08-20")

                @asset(
                    partitions_def=partitions_def
                )
                def upstream_asset():
                    ...

                @asset(
                    partitions_def=partitions_def
                )
                def upstream_asset(context: AssetExecutionContext, upstream_asset):
                    context.log.info(context.asset_partitions_def_for_input("upstream_asset"))

                # materializing the 2023-08-21 partition of this asset will log:
                #   DailyPartitionsDefinition("2023-08-20")

        """
        asset_key = self.asset_key_for_input(input_name)
        result = self._step_execution_context.job_def.asset_layer.partitions_def_for_asset(
            asset_key
        )
        if result is None:
            raise DagsterInvariantViolationError(
                f"Attempting to access partitions def for asset {asset_key}, but it is not"
                " partitioned"
            )

        return result

    @public
    def asset_partition_keys_for_output(self, output_name: str = "result") -> Sequence[str]:
        """Returns a list of the partition keys for the given output.

        If you want to write your asset to support running a backfill of several partitions in a single run,
        you can use ``asset_partition_keys_for_output`` to get all of the partitions being materialized
        by the backfill.

        Args:
            output_name (str): For assets defined with the ``@asset`` decorator, the name of the output
                will be automatically provided. For assets defined with ``@multi_asset``, ``output_name``
                should be the op output associated with the asset key (as determined by AssetOut)
                to get the partition keys for.

        Examples:
            .. code-block:: python

                partitions_def = DailyPartitionsDefinition("2023-08-20")

                @asset(
                    partitions_def=partitions_def
                )
                def an_asset(context: AssetExecutionContext):
                    context.log.info(context.asset_partition_keys_for_output())


                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   ["2023-08-21", "2023-08-22", "2023-08-23", "2023-08-24", "2023-08-25"]

                @multi_asset(
                    outs={
                        "first_asset": AssetOut(key=["my_assets", "first_asset"]),
                        "second_asset": AssetOut(key=["my_assets", "second_asset"])
                    }
                    partitions_def=partitions_def,
                )
                def a_multi_asset(context: AssetExecutionContext):
                    context.log.info(context.asset_partition_keys_for_output("first_asset"))
                    context.log.info(context.asset_partition_keys_for_output("second_asset"))


                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   ["2023-08-21", "2023-08-22", "2023-08-23", "2023-08-24", "2023-08-25"]
                #   ["2023-08-21", "2023-08-22", "2023-08-23", "2023-08-24", "2023-08-25"]


                @asset(
                    partitions_def=partitions_def,
                    ins={
                        "self_dependent_asset": AssetIn(partition_mapping=TimeWindowPartitionMapping(start_offset=-1, end_offset=-1))
                    }
                )
                def self_dependent_asset(context: AssetExecutionContext, self_dependent_asset):
                    context.log.info(context.asset_partition_keys_for_output())

                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   ["2023-08-21", "2023-08-22", "2023-08-23", "2023-08-24", "2023-08-25"]
        """
        return self.asset_partitions_def_for_output(output_name).get_partition_keys_in_range(
            self._step_execution_context.asset_partition_key_range_for_output(output_name),
            dynamic_partitions_store=self.instance,
        )

    @public
    def asset_partition_keys_for_input(self, input_name: str) -> Sequence[str]:
        """Returns a list of the partition keys of the upstream asset corresponding to the
        given input.

        If you want to write your asset to support running a backfill of several partitions in a single run,
        you can use ``asset_partition_keys_for_input`` to get all of the partition keys of the input that
        are relevant to that backfill.

        Args:
            input_name (str): The name of the input to get the time window for.

        Examples:
            .. code-block:: python

                partitions_def = DailyPartitionsDefinition("2023-08-20")

                @asset(
                    partitions_def=partitions_def
                )
                def upstream_asset():
                    ...

                @asset(
                    partitions_def=partitions_def
                )
                def an_asset(context: AssetExecutionContext, upstream_asset):
                    context.log.info(context.asset_partition_keys_for_input("upstream_asset"))


                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   ["2023-08-21", "2023-08-22", "2023-08-23", "2023-08-24", "2023-08-25"]

                @asset(
                    ins={
                        "upstream_asset": AssetIn(partition_mapping=TimeWindowPartitionMapping(start_offset=-1, end_offset=-1))
                    }
                    partitions_def=partitions_def,
                )
                def another_asset(context: AssetExecutionContext, upstream_asset):
                    context.log.info(context.asset_partition_keys_for_input("upstream_asset"))


                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   ["2023-08-20", "2023-08-21", "2023-08-22", "2023-08-23", "2023-08-24"]


                @asset(
                    partitions_def=partitions_def,
                    ins={
                        "self_dependent_asset": AssetIn(partition_mapping=TimeWindowPartitionMapping(start_offset=-1, end_offset=-1))
                    }
                )
                def self_dependent_asset(context: AssetExecutionContext, self_dependent_asset):
                    context.log.info(context.asset_partition_keys_for_input("self_dependent_asset"))

                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   ["2023-08-20", "2023-08-21", "2023-08-22", "2023-08-23", "2023-08-24"]
        """
        return list(
            self._step_execution_context.asset_partitions_subset_for_input(
                input_name
            ).get_partition_keys()
        )

    @public
    def asset_partitions_time_window_for_input(self, input_name: str = "result") -> TimeWindow:
        """The time window for the partitions of the input asset.

        If you want to write your asset to support running a backfill of several partitions in a single run,
        you can use ``asset_partitions_time_window_for_input`` to get the time window of the input that
        are relevant to that backfill.

        Raises an error if either of the following are true:
        - The input asset has no partitioning.
        - The input asset is not partitioned with a TimeWindowPartitionsDefinition or a
        MultiPartitionsDefinition with one time-partitioned dimension.

        Args:
            input_name (str): The name of the input to get the partition key for.

        Examples:
            .. code-block:: python

                partitions_def = DailyPartitionsDefinition("2023-08-20")

                @asset(
                    partitions_def=partitions_def
                )
                def upstream_asset():
                    ...

                @asset(
                    partitions_def=partitions_def
                )
                def an_asset(context: AssetExecutionContext, upstream_asset):
                    context.log.info(context.asset_partitions_time_window_for_input("upstream_asset"))


                # materializing the 2023-08-21 partition of this asset will log:
                #   TimeWindow("2023-08-21", "2023-08-22")

                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   TimeWindow("2023-08-21", "2023-08-26")


                @asset(
                    ins={
                        "upstream_asset": AssetIn(partition_mapping=TimeWindowPartitionMapping(start_offset=-1, end_offset=-1))
                    }
                    partitions_def=partitions_def,
                )
                def another_asset(context: AssetExecutionContext, upstream_asset):
                    context.log.info(context.asset_partitions_time_window_for_input("upstream_asset"))


                # materializing the 2023-08-21 partition of this asset will log:
                #   TimeWindow("2023-08-20", "2023-08-21")

                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   TimeWindow("2023-08-21", "2023-08-26")


                @asset(
                    partitions_def=partitions_def,
                    ins={
                        "self_dependent_asset": AssetIn(partition_mapping=TimeWindowPartitionMapping(start_offset=-1, end_offset=-1))
                    }
                )
                def self_dependent_asset(context: AssetExecutionContext, self_dependent_asset):
                    context.log.info(context.asset_partitions_time_window_for_input("self_dependent_asset"))

                # materializing the 2023-08-21 partition of this asset will log:
                #   TimeWindow("2023-08-20", "2023-08-21")

                # running a backfill of the 2023-08-21 through 2023-08-25 partitions of this asset will log:
                #   TimeWindow("2023-08-20", "2023-08-25")

        """
        return self._step_execution_context.asset_partitions_time_window_for_input(input_name)

    @public
    @experimental
    def get_asset_provenance(self, asset_key: AssetKey) -> Optional[DataProvenance]:
        """Return the provenance information for the most recent materialization of an asset.

        Args:
            asset_key (AssetKey): Key of the asset for which to retrieve provenance.

        Returns:
            Optional[DataProvenance]: Provenance information for the most recent
                materialization of the asset. Returns `None` if the asset was never materialized or
                the materialization record is too old to contain provenance information.
        """
        record = self.instance.get_latest_data_version_record(asset_key)

        return (
            None if record is None else extract_data_provenance_from_entry(record.event_log_entry)
        )

    def set_data_version(self, asset_key: AssetKey, data_version: DataVersion) -> None:
        """Set the data version for an asset being materialized by the currently executing step.
        This is useful for external execution situations where it is not possible to return
        an `Output`.

        Args:
            asset_key (AssetKey): Key of the asset for which to set the data version.
            data_version (DataVersion): The data version to set.
        """
        self._step_execution_context.set_data_version(asset_key, data_version)

    @property
    def asset_check_spec(self) -> AssetCheckSpec:
        asset_checks_def = check.not_none(
            self.job_def.asset_layer.asset_checks_def_for_node(self.node_handle),
            "This context does not correspond to an AssetChecksDefinition",
        )
        return asset_checks_def.spec

    # In this mode no conversion is done on returned values and missing but expected outputs are not
    # allowed.
    @property
    def requires_typed_event_stream(self) -> bool:
        return self._step_execution_context.requires_typed_event_stream

    @property
    def typed_event_stream_error_message(self) -> Optional[str]:
        return self._step_execution_context.typed_event_stream_error_message

    def set_requires_typed_event_stream(self, *, error_message: Optional[str] = None) -> None:
        self._step_execution_context.set_requires_typed_event_stream(error_message=error_message)


class AssetExecutionContext(OpExecutionContext):
    def __init__(self, step_execution_context: StepExecutionContext):
        super().__init__(step_execution_context=step_execution_context)


def build_execution_context(
    step_context: StepExecutionContext,
) -> Union[OpExecutionContext, AssetExecutionContext]:
    """Get the correct context based on the type of step (op or asset) and the user provided context
    type annotation. Follows these rules.

    step type     annotation                result
    asset         AssetExecutionContext     AssetExecutionContext
    asset         OpExecutionContext        OpExecutionContext
    asset         None                      AssetExecutionContext
    op            AssetExecutionContext     Error - we cannot init an AssetExecutionContext w/o an AssetsDefinition
    op            OpExecutionContext        OpExecutionContext
    op            None                      OpExecutionContext
    asset_check   AssetExecutionContext     AssetExecutionContext
    asset_check   OpExecutionContext        OpExecutionContext
    asset_check   None                      AssetExecutionContext

    For ops in graph-backed assets
    step type     annotation                result
    op            AssetExecutionContext     AssetExecutionContext
    op            OpExecutionContext        OpExecutionContext
    op            None                      OpExecutionContext
    """
    is_sda_step = step_context.is_sda_step
    is_op_in_graph_asset = step_context.is_in_graph_asset
    is_asset_check = step_context.is_asset_check_step
    context_annotation = EmptyAnnotation
    compute_fn = step_context.op_def._compute_fn  # noqa: SLF001
    compute_fn = (
        compute_fn
        if isinstance(compute_fn, DecoratedOpFunction)
        else DecoratedOpFunction(compute_fn)
    )
    if compute_fn.has_context_arg():
        context_param = compute_fn.get_context_arg()
        context_annotation = context_param.annotation

    # It would be nice to do this check at definition time, rather than at run time, but we don't
    # know if the op is part of an op job or a graph-backed asset until we have the step execution context
    if (
        context_annotation is AssetExecutionContext
        and not is_sda_step
        and not is_asset_check
        and not is_op_in_graph_asset
    ):
        # AssetExecutionContext requires an AssetsDefinition during init, so an op in an op job
        # cannot be annotated with AssetExecutionContext
        raise DagsterInvalidDefinitionError(
            "Cannot annotate @op `context` parameter with type AssetExecutionContext unless the"
            " op is part of a graph-backed asset. `context` must be annotated with"
            " OpExecutionContext, or left blank."
        )

    if context_annotation is EmptyAnnotation:
        # if no type hint has been given, default to:
        # * AssetExecutionContext for sda steps not in graph-backed assets, and asset_checks
        # * OpExecutionContext for non sda steps
        # * OpExecutionContext for ops in graph-backed assets
        if is_asset_check:
            return AssetExecutionContext(step_context)
        if is_op_in_graph_asset or not is_sda_step:
            return OpExecutionContext(step_context)
        return AssetExecutionContext(step_context)
    if context_annotation is AssetExecutionContext:
        return AssetExecutionContext(step_context)
    return OpExecutionContext(step_context)
